
#Import "toybox/toybox"
#Import "<mojo3d-vr>"

Using std..
Using mojo..
Using mojo3d..
Using openvr..

Using ToyBox..

Class VRCamera Extends Camera

	Field pivot:Model
	
	' Give Camera a parent pivot to allow free movement of
	' camera while including user's own head movements...
	
	Method New ()
		pivot	= New Model
		Parent	= pivot
	End

	' Update camera's own local matrix from VRRenderer's headset matrix...
	
	Method Update (renderer:VRRenderer)
		LocalBasis		= renderer.HeadMatrix.m
		LocalPosition	= renderer.HeadMatrix.t
	End

	' Move parent pivot instead of camera...
	
	Method MoveVR (tx:Float, ty:Float, tz:Float) ' Yuk!
		pivot.Move (tx, ty, tz)
	End

	Method RotateVR (rx:Float, ry:Float, rz:Float, localSpace:Bool = False)
		pivot.Rotate (rx, ry, rz, localSpace)
	End
End

Class Canvas Extension

	Method ShadowText (text:String, tx:Float, ty:Float, handleX:Float = 0.0, handleY:Float = 0.0)
		Color = Color.Black
		DrawText (text, tx + 1, ty + 1, handleX, handleY)
		Color = Color.White
		DrawText (text, tx, ty, handleX, handleY)
	End
	
End

Class Game Extends Window

	Const CAMERA_SPEED_NORMAL:Float	= 0.1
	Const CAMERA_SPEED_TURBO:Float	= 1.0

	Field boost_camera:Bool	' Camera speed boost (left-shift sets True)
	
	Field scene:Scene
	Field camera:VRCamera
	Field light:Light

	Field renderer:VRRenderer
	
	Field center:Model

	' Wall-specific demo stuff...

	Field wall_left:BrickWall
	Field wall_right:BrickWall
	Field wall_front:BrickWall
	Field wall_back:BrickWall
	
	Method New (title:String, width:Int, height:Int, flags:WindowFlags)
		
		Super.New (title, width, height, flags)
		
		renderer			= New VRRenderer

			renderer.TrackingSpace = VRTrackingSpace.Seated
		
		scene = Scene.GetCurrent ()

			scene.CSMSplits	= New Float [] (2, 4, 16, 256)
		
		center = New Model
		
			center.Move (0, 15, 0)
		
		camera = New VRCamera

			camera.Near	= 0.1
			camera.Far	= 500
			
			camera.MoveVR (0, 5, -30)
			camera.PointAt (center)
			
		light = New Light
			
			light.Type = LightType.Directional
			light.CastsShadow = False
			
			light.Move (-100, 100, -100)
			light.PointAt (center)

		Local ground:GroundBody = New GroundBody (, , , New PbrMaterial (Color.Green * 0.3))
		
			ground.Start ()

		BuildBrickShotHouse (4, 8, 2, 1, 1)
		
	End

	Method BuildBrickShotHouse (wall_width:Float = 5, wall_height:Float = 50, wall_brick_width:Float = 5, wall_brick_height:Float = 2.5, wall_brick_depth:Float = 2.5)

			Local size:Float = wall_width * wall_brick_width
			Local half:Float = size * 0.5
			
			Local half_brick_width:Float = wall_brick_width * 0.5
			
			wall_front = New BrickWall (wall_width, wall_height, wall_brick_width, wall_brick_height, wall_brick_depth)
		
				wall_front.Start ()
		
			wall_back = New BrickWall (wall_width, wall_height, wall_brick_width, wall_brick_height, wall_brick_depth)
		
				wall_back.Move (-wall_brick_width, 0, size)
				wall_back.Rotate (0, 180, 0)
				wall_back.Start ()
		
			wall_left = New BrickWall (wall_width, wall_height, wall_brick_width, wall_brick_height, wall_brick_depth)
		
				wall_left.Move (-(half + half_brick_width), 0, half - half_brick_width)
				wall_left.Rotate (0, -90, 0)
				wall_left.Start ()
		
			wall_right= New BrickWall (wall_width, wall_height, wall_brick_width, wall_brick_height, wall_brick_depth)
		
				wall_right.Move (half - half_brick_width, 0, half + half_brick_width)
				wall_right.Rotate (0, 90, 0)
				wall_right.Start ()

	End
	
	Method ProcessInput:Void ()

		If Keyboard.KeyHit (Key.Enter)
			renderer.ResetSeatedPose ()
		Endif

		If Keyboard.KeyHit (Key.F5)
		
			For Local b:BoxBody = Eachin ToyBox.BoxList
				b.model.Destroy ()
				ToyBox.BoxList.Remove (b)
			Next
			
			For Local b:SphereBody = Eachin ToyBox.SphereList
				b.model.Destroy ()
				ToyBox.SphereList.Remove (b)
			Next
			
			For Local b:CapsuleBody = Eachin ToyBox.CapsuleList
				b.model.Destroy ()
				ToyBox.CapsuleList.Remove (b)
			Next
			
			For Local b:ConeBody = Eachin ToyBox.ConeList
				b.model.Destroy ()
				ToyBox.ConeList.Remove (b)
			Next
			
			For Local b:CylinderBody = Eachin ToyBox.CylinderList
				b.model.Destroy ()
				ToyBox.CylinderList.Remove (b)
			Next
			
			BuildBrickShotHouse ()
		
		Endif

		If Keyboard.KeyHit (Key.Escape)
			VR_Shutdown ()
			App.Terminate ()
		Endif
		
		' Shoot from camera position with force, size, mass:
		
		If Keyboard.KeyHit (Key.Space) Then Shoot (camera, 100.0, 2.0, 1.0)
		
		If Keyboard.KeyHit (Key.S) Then light.CastsShadow = Not light.CastsShadow

		If Keyboard.KeyDown (Key.LeftShift)
			boost_camera = True
		Else
			boost_camera = False
		Endif
		
		If Keyboard.KeyDown (Key.A)
			If boost_camera
				camera.MoveVR (0.0, 0.0, CAMERA_SPEED_TURBO)
			Else
				camera.MoveVR (0.0, 0.0, CAMERA_SPEED_NORMAL)
			Endif
		Endif
 
		If Keyboard.KeyDown (Key.Z)
			If boost_camera
				camera.MoveVR (0.0, 0.0, -CAMERA_SPEED_TURBO)
			Else
				camera.MoveVR (0.0, 0.0, -CAMERA_SPEED_NORMAL)
			Endif
		Endif
 
		If Keyboard.KeyDown (Key.Left)
			camera.RotateVR (0.0, 1.0, 0.0)
		Endif
 
		If Keyboard.KeyDown (Key.Right)
			camera.RotateVR (0.0, -1.0, 0.0)
		Endif
 
		If Keyboard.KeyDown (Key.Up)
			camera.RotateVR (1.0, 0.0, 0.0, True)
		Endif
 
		If Keyboard.KeyDown (Key.Down)
			camera.RotateVR (-1.0, 0.0, 0.0, True)
		Endif
		
		camera.FOV = camera.FOV - (Mouse.WheelY * 5.0)
		
	End
	
	Method UpdateGame:Void ()

		ProcessInput ()
		scene.Update ()

	End
	
	Method OnRender (canvas:Canvas) Override
 
 		UpdateGame ()
		RequestRender ()

		renderer.Update ()						' Update the VR renderer...
		
		camera.Update (renderer) ' camera.Matrix = renderer.HeadMatrix		' Set the camera to the user's head position...

		scene.Render (canvas)

		' ShadowText is in class extension at top!
		
		canvas.ShadowText ("FPS: " + App.FPS,									20.0, 20.0)
		
		canvas.ShadowText ("A/Z + cursors to move; left-shift to boost",		20.0, 50.0)
		canvas.ShadowText ("Space to shoot!",									20.0, 70.0)
		canvas.ShadowText ("D to drop random bodies; F5 to reset scene",		20.0, 90.0)
		canvas.ShadowText ("S to toggle shadows",								20.0, 110.0)
		
		canvas.ShadowText ("Boxes: "		+ ToyBox.BoxList.Count (),		20.0, 150.0)
		canvas.ShadowText ("Spheres: "	+ ToyBox.SphereList.Count (),	20.0, 170.0)
		canvas.ShadowText ("Capsules: "	+ ToyBox.CapsuleList.Count (),	20.0, 190.0)
		canvas.ShadowText ("Cones: "		+ ToyBox.ConeList.Count (),		20.0, 210.0)

		canvas.Color = Color.Black
		canvas.DrawRect (Width * 0.5, Height * 0.5, 2.0, 2.0)

		canvas.Color = Color.White
		canvas.DrawRect (Width * 0.5 - 1.0, Height * 0.5 - 1.0, 2.0, 2.0)

	End

End

Function Main ()
 
	Local width:Int			= 640
	Local height:Int		= 480
	Local flags:WindowFlags	= WindowFlags.Center
	
'	Local width:Int			= 1920
'	Local height:Int		= 1080
'	Local flags:WindowFlags	= WindowFlags.Fullscreen
	
	New AppInstance
	
	New Game ("Physics Demo", width, height, flags)
	
	App.Run ()
	
End
